@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Distribution.Client.IndexUtils
@@ -11,14 +12,19 @@
 -- Extra utils related to the package indexes.
 -----------------------------------------------------------------------------
 module Distribution.Client.IndexUtils (
+  getIndexFileAge,
   getInstalledPackages,
   getSourcePackages,
+  getSourcePackagesStrict,
   convert,
 
   readPackageIndexFile,
   parsePackageIndex,
   readRepoIndex,
   updateRepoIndexCache,
+  updatePackageIndexCacheFile,
+
+  BuildTreeRefType(..), refTypeFromTypeCode, typeCodeFromRefType
   ) where
 
 import qualified Distribution.Client.Tar as Tar
@@ -29,9 +35,11 @@ import Distribution.Package
          , Package(..), packageVersion, packageName
          , Dependency(Dependency), InstalledPackageId(..) )
 import Distribution.Client.PackageIndex (PackageIndex)
-import qualified Distribution.Client.PackageIndex as PackageIndex
-import qualified Distribution.Simple.PackageIndex as InstalledPackageIndex
-import qualified Distribution.InstalledPackageInfo as InstalledPackageInfo
+import qualified Distribution.Client.PackageIndex      as PackageIndex
+import Distribution.Simple.PackageIndex (InstalledPackageIndex)
+import qualified Distribution.Simple.PackageIndex      as InstalledPackageIndex
+import qualified Distribution.InstalledPackageInfo     as InstalledPackageInfo
+import qualified Distribution.PackageDescription.Parse as PackageDesc.Parse
 import Distribution.PackageDescription
          ( GenericPackageDescription )
 import Distribution.PackageDescription.Parse
@@ -49,48 +57,51 @@ import Distribution.Version
 import Distribution.Text
          ( display, simpleParse )
 import Distribution.Verbosity
-         ( Verbosity, lessVerbose )
+         ( Verbosity, normal, lessVerbose )
 import Distribution.Simple.Utils
          ( die, warn, info, fromUTF8 )
 
 import Data.Char   (isAlphaNum)
-import Data.Maybe  (catMaybes, fromMaybe)
+import Data.Maybe  (mapMaybe, fromMaybe)
 import Data.List   (isPrefixOf)
+#if !MIN_VERSION_base(4,8,0)
 import Data.Monoid (Monoid(..))
+#endif
 import qualified Data.Map as Map
-import Control.Monad (MonadPlus(mplus), when, unless, liftM)
+import Control.Monad (MonadPlus(mplus), when, liftM)
 import Control.Exception (evaluate)
 import qualified Data.ByteString.Lazy as BS
 import qualified Data.ByteString.Lazy.Char8 as BS.Char8
 import qualified Data.ByteString.Char8 as BSS
 import Data.ByteString.Lazy (ByteString)
 import Distribution.Client.GZipUtils (maybeDecompress)
+import Distribution.Client.Utils ( byteStringToFilePath
+                                 , tryFindAddSourcePackageDesc )
+import Distribution.Compat.Exception (catchIO)
+import Distribution.Client.Compat.Time (getFileAge, getModTime)
+import System.Directory (doesFileExist)
 import System.FilePath ((</>), takeExtension, splitDirectories, normalise)
 import System.FilePath.Posix as FilePath.Posix
          ( takeFileName )
 import System.IO
 import System.IO.Unsafe (unsafeInterleaveIO)
 import System.IO.Error (isDoesNotExistError)
-import Distribution.Compat.Exception (catchIO)
-import System.Directory
-         ( getModificationTime, doesFileExist )
-import System.Time
-         ( getClockTime, diffClockTimes, normalizeTimeDiff, TimeDiff(tdDay) )
+import Numeric (showFFloat)
 
 
 getInstalledPackages :: Verbosity -> Compiler
                      -> PackageDBStack -> ProgramConfiguration
-                     -> IO InstalledPackageIndex.PackageIndex
+                     -> IO InstalledPackageIndex
 getInstalledPackages verbosity comp packageDbs conf =
     Configure.getInstalledPackages verbosity' comp packageDbs conf
   where
     --FIXME: make getInstalledPackages use sensible verbosity in the first place
     verbosity'  = lessVerbose verbosity
 
-convert :: InstalledPackageIndex.PackageIndex -> PackageIndex InstalledPackage
+convert :: InstalledPackageIndex -> PackageIndex InstalledPackage
 convert index' = PackageIndex.fromList
     -- There can be multiple installed instances of each package version,
-    -- like when the same package is installed in the global & user dbs.
+    -- like when the same package is installed in the global & user DBs.
     -- InstalledPackageIndex.allPackagesBySourcePackageId gives us the
     -- installed packages with the most preferred instances first, so by
     -- picking the first we should get the user one. This is almost but not
@@ -106,7 +117,7 @@ convert index' = PackageIndex.fromList
     -- other packages that do not exist then we have a problem we cannot find
     -- the original source package id. Instead we make up a bogus package id.
     -- This should have the same effect since it should be a dependency on a
-    -- non-existant package.
+    -- nonexistent package.
     sourceDeps index ipkg =
       [ maybe (brokenPackageId depid) packageId mdep
       | let depids = InstalledPackageInfo.depends ipkg
@@ -129,16 +140,29 @@ convert index' = PackageIndex.fromList
 -- This is a higher level wrapper used internally in cabal-install.
 --
 getSourcePackages :: Verbosity -> [Repo] -> IO SourcePackageDb
-getSourcePackages verbosity [] = do
+getSourcePackages verbosity repos = getSourcePackages' verbosity repos
+                                    ReadPackageIndexLazyIO
+
+-- | Like 'getSourcePackages', but reads the package index strictly. Useful if
+-- you want to write to the package index after having read it.
+getSourcePackagesStrict :: Verbosity -> [Repo] -> IO SourcePackageDb
+getSourcePackagesStrict verbosity repos = getSourcePackages' verbosity repos
+                                          ReadPackageIndexStrict
+
+-- | Common implementation used by getSourcePackages and
+-- getSourcePackagesStrict.
+getSourcePackages' :: Verbosity -> [Repo] -> ReadPackageIndexMode
+                      -> IO SourcePackageDb
+getSourcePackages' verbosity [] _mode = do
   warn verbosity $ "No remote package servers have been specified. Usually "
                 ++ "you would have one specified in the config file."
   return SourcePackageDb {
     packageIndex       = mempty,
     packagePreferences = mempty
   }
-getSourcePackages verbosity repos = do
+getSourcePackages' verbosity repos mode = do
   info verbosity "Reading available packages..."
-  pkgss <- mapM (readRepoIndex verbosity) repos
+  pkgss <- mapM (\r -> readRepoIndex verbosity r mode) repos
   let (pkgs, prefs) = mconcat pkgss
       prefs' = Map.fromListWith intersectVersionRanges
                  [ (name, range) | Dependency name range <- prefs ]
@@ -149,7 +173,6 @@ getSourcePackages verbosity repos = do
     packagePreferences = prefs'
   }
 
-
 -- | Read a repository index from disk, from the local file specified by
 -- the 'Repo'.
 --
@@ -157,25 +180,31 @@ getSourcePackages verbosity repos = do
 --
 -- This is a higher level wrapper used internally in cabal-install.
 --
-readRepoIndex :: Verbosity -> Repo
+readRepoIndex :: Verbosity -> Repo -> ReadPackageIndexMode
               -> IO (PackageIndex SourcePackage, [Dependency])
-readRepoIndex verbosity repo =
+readRepoIndex verbosity repo mode =
   let indexFile = repoLocalDir repo </> "00-index.tar"
       cacheFile = repoLocalDir repo </> "00-index.cache"
   in handleNotFound $ do
-    warnIfIndexIsOld indexFile
+    warnIfIndexIsOld =<< getIndexFileAge repo
     whenCacheOutOfDate indexFile cacheFile $ do
-      info verbosity $ "Updating the index cache file..."
-      updatePackageIndexCacheFile indexFile cacheFile
-    readPackageIndexCacheFile mkAvailablePackage indexFile cacheFile
+      updatePackageIndexCacheFile verbosity indexFile cacheFile
+    readPackageIndexCacheFile mkAvailablePackage indexFile cacheFile mode
 
   where
-    mkAvailablePackage pkgid pkg =
+    mkAvailablePackage pkgEntry =
       SourcePackage {
         packageInfoId      = pkgid,
-        packageDescription = pkg,
-        packageSource      = RepoTarballPackage repo pkgid Nothing
+        packageDescription = packageDesc pkgEntry,
+        packageSource      = case pkgEntry of
+          NormalPackage _ _ _ _       -> RepoTarballPackage repo pkgid Nothing
+          BuildTreeRef  _  _ _ path _ -> LocalUnpackedPackage path,
+        packageDescrOverride = case pkgEntry of
+          NormalPackage _ _ pkgtxt _ -> Just pkgtxt
+          _                          -> Nothing
       }
+      where
+        pkgid = packageId pkgEntry
 
     handleNotFound action = catchIO action $ \e -> if isDoesNotExistError e
       then do
@@ -190,74 +219,104 @@ readRepoIndex verbosity repo =
       else ioError e
 
     isOldThreshold = 15 --days
-    warnIfIndexIsOld indexFile = do
-      indexTime   <- getModificationTime indexFile
-      currentTime <- getClockTime
-      let diff = normalizeTimeDiff (diffClockTimes currentTime indexTime)
-      when (tdDay diff >= isOldThreshold) $ case repoKind repo of
+    warnIfIndexIsOld dt = do
+      when (dt >= isOldThreshold) $ case repoKind repo of
         Left  remoteRepo -> warn verbosity $
              "The package list for '" ++ remoteRepoName remoteRepo
-          ++ "' is " ++ show (tdDay diff)  ++ " days old.\nRun "
+          ++ "' is " ++ showFFloat (Just 1) dt " days old.\nRun "
           ++ "'cabal update' to get the latest list of available packages."
         Right _localRepo -> return ()
 
+
+-- | Return the age of the index file in days (as a Double).
+getIndexFileAge :: Repo -> IO Double
+getIndexFileAge repo = getFileAge $ repoLocalDir repo </> "00-index.tar"
+
+
 -- | It is not necessary to call this, as the cache will be updated when the
 -- index is read normally. However you can do the work earlier if you like.
 --
 updateRepoIndexCache :: Verbosity -> Repo -> IO ()
 updateRepoIndexCache verbosity repo =
     whenCacheOutOfDate indexFile cacheFile $ do
-      info verbosity $ "Updating the index cache file..."
-      updatePackageIndexCacheFile indexFile cacheFile
+      updatePackageIndexCacheFile verbosity indexFile cacheFile
   where
     indexFile = repoLocalDir repo </> "00-index.tar"
     cacheFile = repoLocalDir repo </> "00-index.cache"
 
-whenCacheOutOfDate :: FilePath-> FilePath -> IO () -> IO ()
+whenCacheOutOfDate :: FilePath -> FilePath -> IO () -> IO ()
 whenCacheOutOfDate origFile cacheFile action = do
   exists <- doesFileExist cacheFile
   if not exists
     then action
     else do
-      origTime  <- getModificationTime origFile
-      cacheTime <- getModificationTime cacheFile
-      unless (cacheTime >= origTime) action
-
+      origTime  <- getModTime origFile
+      cacheTime <- getModTime cacheFile
+      when (origTime > cacheTime) action
 
 ------------------------------------------------------------------------
 -- Reading the index file
 --
 
+-- | An index entry is either a normal package, or a local build tree reference.
+data PackageEntry =
+  NormalPackage  PackageId GenericPackageDescription ByteString BlockNo
+  | BuildTreeRef BuildTreeRefType
+                 PackageId GenericPackageDescription FilePath   BlockNo
+
+-- | A build tree reference is either a link or a snapshot.
+data BuildTreeRefType = SnapshotRef | LinkRef
+                      deriving Eq
+
+refTypeFromTypeCode :: Tar.TypeCode -> BuildTreeRefType
+refTypeFromTypeCode t
+  | t == Tar.buildTreeRefTypeCode      = LinkRef
+  | t == Tar.buildTreeSnapshotTypeCode = SnapshotRef
+  | otherwise                          =
+    error "Distribution.Client.IndexUtils.refTypeFromTypeCode: unknown type code"
+
+typeCodeFromRefType :: BuildTreeRefType -> Tar.TypeCode
+typeCodeFromRefType LinkRef     = Tar.buildTreeRefTypeCode
+typeCodeFromRefType SnapshotRef = Tar.buildTreeSnapshotTypeCode
+
+type MkPackageEntry = IO PackageEntry
+
+instance Package PackageEntry where
+  packageId (NormalPackage  pkgid _ _ _) = pkgid
+  packageId (BuildTreeRef _ pkgid _ _ _) = pkgid
+
+packageDesc :: PackageEntry -> GenericPackageDescription
+packageDesc (NormalPackage  _ descr _ _) = descr
+packageDesc (BuildTreeRef _ _ descr _ _) = descr
+
 -- | Read a compressed \"00-index.tar.gz\" file into a 'PackageIndex'.
 --
 -- This is supposed to be an \"all in one\" way to easily get at the info in
--- the hackage package index.
+-- the Hackage package index.
 --
 -- It takes a function to map a 'GenericPackageDescription' into any more
 -- specific instance of 'Package' that you might want to use. In the simple
 -- case you can just use @\_ p -> p@ here.
 --
 readPackageIndexFile :: Package pkg
-                     => (PackageId -> GenericPackageDescription -> pkg)
+                     => (PackageEntry -> pkg)
                      -> FilePath
                      -> IO (PackageIndex pkg, [Dependency])
 readPackageIndexFile mkPkg indexFile = do
-  (pkgs, prefs) <- either fail return
-                 . parsePackageIndex
-                 . maybeDecompress
-               =<< BS.readFile indexFile
+  (mkPkgs, prefs) <- either fail return
+                     . parsePackageIndex
+                     . maybeDecompress
+                     =<< BS.readFile indexFile
 
-  pkgs' <- evaluate $ PackageIndex.fromList
-            [ mkPkg pkgid pkg | (pkgid, pkg, _) <- pkgs]
-  return (pkgs', prefs)
+  pkgEntries  <- sequence mkPkgs
+  pkgs <- evaluate $ PackageIndex.fromList (map mkPkg pkgEntries)
+  return (pkgs, prefs)
 
 -- | Parse an uncompressed \"00-index.tar\" repository index file represented
 -- as a 'ByteString'.
 --
 parsePackageIndex :: ByteString
-                  -> Either String
-                            ( [(PackageId, GenericPackageDescription, BlockNo)]
-                            , [Dependency] )
+                  -> Either String ([MkPackageEntry], [Dependency])
 parsePackageIndex = accum 0 [] [] . Tar.read
   where
     accum blockNo pkgs prefs es = case es of
@@ -266,7 +325,7 @@ parsePackageIndex = accum 0 [] [] . Tar.read
       Tar.Next e es' -> accum blockNo' pkgs' prefs' es'
         where
           (pkgs', prefs') = extract blockNo pkgs prefs e
-          blockNo'        = blockNo + sizeInBlocks e
+          blockNo'        = blockNo + Tar.entrySizeInBlocks e
 
     extract blockNo pkgs prefs entry =
        fromMaybe (pkgs, prefs) $
@@ -274,28 +333,20 @@ parsePackageIndex = accum 0 [] [] . Tar.read
          `mplus` tryExtractPrefs
       where
         tryExtractPkg = do
-          (pkgid, pkg) <- extractPkg entry
-          return ((pkgid, pkg, blockNo):pkgs, prefs)
+          mkPkgEntry <- extractPkg entry blockNo
+          return (mkPkgEntry:pkgs, prefs)
 
         tryExtractPrefs = do
           prefs' <- extractPrefs entry
           return (pkgs, prefs'++prefs)
 
-    sizeInBlocks entry =
-        1 + case Tar.entryContent entry of
-              Tar.NormalFile     _   size -> bytesToBlocks size
-              Tar.OtherEntryType _ _ size -> bytesToBlocks size
-              _                           -> 0
-      where
-        bytesToBlocks s = 1 + ((fromIntegral s - 1) `div` 512)
-
-extractPkg :: Tar.Entry -> Maybe (PackageId, GenericPackageDescription)
-extractPkg entry = case Tar.entryContent entry of
+extractPkg :: Tar.Entry -> BlockNo -> Maybe MkPackageEntry
+extractPkg entry blockNo = case Tar.entryContent entry of
   Tar.NormalFile content _
      | takeExtension fileName == ".cabal"
     -> case splitDirectories (normalise fileName) of
         [pkgname,vers,_] -> case simpleParse vers of
-          Just ver -> Just (pkgid, descr)
+          Just ver -> Just $ return (NormalPackage pkgid descr content blockNo)
             where
               pkgid  = PackageIdentifier (PackageName pkgname) ver
               parsed = parsePackageDescription . fromUTF8 . BS.Char8.unpack
@@ -306,7 +357,19 @@ extractPkg entry = case Tar.entryContent entry of
                                     ++ show fileName
           _ -> Nothing
         _ -> Nothing
+
+  Tar.OtherEntryType typeCode content _
+    | Tar.isBuildTreeRefTypeCode typeCode ->
+      Just $ do
+        let path   = byteStringToFilePath content
+            err = "Error reading package index."
+        cabalFile <- tryFindAddSourcePackageDesc path err
+        descr     <- PackageDesc.Parse.readPackageDescription normal cabalFile
+        return $ BuildTreeRef (refTypeFromTypeCode typeCode) (packageId descr)
+                              descr path blockNo
+
   _ -> Nothing
+
   where
     fileName = Tar.entryPath entry
 
@@ -319,8 +382,7 @@ extractPrefs entry = case Tar.entryContent entry of
   _ -> Nothing
 
 parsePreferredVersions :: String -> [Dependency]
-parsePreferredVersions = catMaybes
-                       . map simpleParse
+parsePreferredVersions = mapMaybe simpleParse
                        . filter (not . isPrefixOf "--")
                        . lines
 
@@ -328,36 +390,51 @@ parsePreferredVersions = catMaybes
 -- Reading and updating the index cache
 --
 
-updatePackageIndexCacheFile :: FilePath -> FilePath -> IO ()
-updatePackageIndexCacheFile indexFile cacheFile = do
-    (pkgs, prefs) <- either fail return
-                   . parsePackageIndex
-                   . maybeDecompress
-                 =<< BS.readFile indexFile
-    let cache = mkCache pkgs prefs
+updatePackageIndexCacheFile :: Verbosity -> FilePath -> FilePath -> IO ()
+updatePackageIndexCacheFile verbosity indexFile cacheFile = do
+    info verbosity "Updating the index cache file..."
+    (mkPkgs, prefs) <- either fail return
+                       . parsePackageIndex
+                       . maybeDecompress
+                       =<< BS.readFile indexFile
+    pkgEntries <- sequence mkPkgs
+    let cache = mkCache pkgEntries prefs
     writeFile cacheFile (showIndexCache cache)
   where
     mkCache pkgs prefs =
         [ CachePreference pref          | pref <- prefs ]
-     ++ [ CachePackageId pkgid blockNo | (pkgid, _, blockNo) <- pkgs ]
+     ++ [ CachePackageId pkgid blockNo
+        | (NormalPackage pkgid _ _ blockNo) <- pkgs ]
+     ++ [ CacheBuildTreeRef refType blockNo
+        | (BuildTreeRef refType _ _ _ blockNo) <- pkgs]
+
+data ReadPackageIndexMode = ReadPackageIndexStrict
+                          | ReadPackageIndexLazyIO
 
 readPackageIndexCacheFile :: Package pkg
-                          => (PackageId -> GenericPackageDescription -> pkg)
+                          => (PackageEntry -> pkg)
                           -> FilePath
                           -> FilePath
+                          -> ReadPackageIndexMode
                           -> IO (PackageIndex pkg, [Dependency])
-readPackageIndexCacheFile mkPkg indexFile cacheFile = do
-  indexHnd <- openFile indexFile ReadMode
+readPackageIndexCacheFile mkPkg indexFile cacheFile mode = do
   cache    <- liftM readIndexCache (BSS.readFile cacheFile)
-  packageIndexFromCache mkPkg indexHnd cache
+  myWithFile indexFile ReadMode $ \indexHnd ->
+    packageIndexFromCache mkPkg indexHnd cache mode
+  where
+    myWithFile f m act = case mode of
+      ReadPackageIndexStrict -> withFile f m act
+      ReadPackageIndexLazyIO -> do indexHnd <- openFile f m
+                                   act indexHnd
 
 
 packageIndexFromCache :: Package pkg
-                      => (PackageId -> GenericPackageDescription -> pkg)
+                      => (PackageEntry -> pkg)
                       -> Handle
                       -> [IndexCacheEntry]
+                      -> ReadPackageIndexMode
                       -> IO (PackageIndex pkg, [Dependency])
-packageIndexFromCache mkPkg hnd = accum mempty []
+packageIndexFromCache mkPkg hnd entrs mode = accum mempty [] entrs
   where
     accum srcpkgs prefs [] = do
       -- Have to reverse entries, since in a tar file, later entries mask
@@ -371,29 +448,52 @@ packageIndexFromCache mkPkg hnd = accum mempty []
       -- The magic here is that we use lazy IO to read the .cabal file
       -- from the index tarball if it turns out that we need it.
       -- Most of the time we only need the package id.
-      pkg <- unsafeInterleaveIO $ do
-               getPackageDescription blockno
-      let srcpkg = mkPkg pkgid pkg
+      ~(pkg, pkgtxt) <- unsafeInterleaveIO $ do
+        pkgtxt <- getEntryContent blockno
+        pkg    <- readPackageDescription pkgtxt
+        return (pkg, pkgtxt)
+      let srcpkg = case mode of
+            ReadPackageIndexLazyIO ->
+              mkPkg (NormalPackage pkgid pkg pkgtxt blockno)
+            ReadPackageIndexStrict ->
+              pkg `seq` pkgtxt `seq` mkPkg (NormalPackage pkgid pkg
+                                            pkgtxt blockno)
+      accum (srcpkg:srcpkgs) prefs entries
+
+    accum srcpkgs prefs (CacheBuildTreeRef refType blockno : entries) = do
+      -- We have to read the .cabal file eagerly here because we can't cache the
+      -- package id for build tree references - the user might edit the .cabal
+      -- file after the reference was added to the index.
+      path <- liftM byteStringToFilePath . getEntryContent $ blockno
+      pkg  <- do let err = "Error reading package index from cache."
+                 file <- tryFindAddSourcePackageDesc path err
+                 PackageDesc.Parse.readPackageDescription normal file
+      let srcpkg = mkPkg (BuildTreeRef refType (packageId pkg) pkg path blockno)
       accum (srcpkg:srcpkgs) prefs entries
 
     accum srcpkgs prefs (CachePreference pref : entries) =
       accum srcpkgs (pref:prefs) entries
 
-    getPackageDescription blockno = do
+    getEntryContent :: BlockNo -> IO ByteString
+    getEntryContent blockno = do
       hSeek hnd AbsoluteSeek (fromIntegral (blockno * 512))
       header  <- BS.hGet hnd 512
       size    <- getEntrySize header
-      content <- BS.hGet hnd (fromIntegral size)
-      readPackageDescription content
+      BS.hGet hnd (fromIntegral size)
 
+    getEntrySize :: ByteString -> IO Tar.FileSize
     getEntrySize header =
       case Tar.read header of
         Tar.Next e _ ->
           case Tar.entryContent e of
             Tar.NormalFile _ size -> return size
+            Tar.OtherEntryType typecode _ size
+              | Tar.isBuildTreeRefTypeCode typecode
+                                  -> return size
             _                     -> interror "unexpected tar entry type"
         _ -> interror "could not read tar file entry"
 
+    readPackageDescription :: ByteString -> IO GenericPackageDescription
     readPackageDescription content =
       case parsePackageDescription . fromUTF8 . BS.Char8.unpack $ content of
         ParseOk _ d -> return d
@@ -413,27 +513,36 @@ packageIndexFromCache mkPkg hnd = accum mempty []
 type BlockNo = Int
 
 data IndexCacheEntry = CachePackageId PackageId BlockNo
+                     | CacheBuildTreeRef BuildTreeRefType BlockNo
                      | CachePreference Dependency
-  deriving (Eq, Show)
+  deriving (Eq)
+
+packageKey, blocknoKey, buildTreeRefKey, preferredVersionKey :: String
+packageKey = "pkg:"
+blocknoKey = "b#"
+buildTreeRefKey     = "build-tree-ref:"
+preferredVersionKey = "pref-ver:"
 
 readIndexCacheEntry :: BSS.ByteString -> Maybe IndexCacheEntry
 readIndexCacheEntry = \line ->
   case BSS.words line of
     [key, pkgnamestr, pkgverstr, sep, blocknostr]
-      | key == packageKey && sep == blocknoKey ->
+      | key == BSS.pack packageKey && sep == BSS.pack blocknoKey ->
       case (parseName pkgnamestr, parseVer pkgverstr [],
             parseBlockNo blocknostr) of
         (Just pkgname, Just pkgver, Just blockno)
           -> Just (CachePackageId (PackageIdentifier pkgname pkgver) blockno)
         _ -> Nothing
-    (key: remainder) | key == preferredVersionKey ->
+    [key, typecodestr, blocknostr] | key == BSS.pack buildTreeRefKey ->
+      case (parseRefType typecodestr, parseBlockNo blocknostr) of
+        (Just refType, Just blockno)
+          -> Just (CacheBuildTreeRef refType blockno)
+        _ -> Nothing
+
+    (key: remainder) | key == BSS.pack preferredVersionKey ->
       fmap CachePreference (simpleParse (BSS.unpack (BSS.unwords remainder)))
     _  -> Nothing
   where
-    packageKey = BSS.pack "pkg:"
-    blocknoKey = BSS.pack "b#"
-    preferredVersionKey = BSS.pack "pref-ver:"
-
     parseName str
       | BSS.all (\c -> isAlphaNum c || c == '-') str
                   = Just (PackageName (BSS.unpack str))
@@ -452,15 +561,31 @@ readIndexCacheEntry = \line ->
         Just (blockno, remainder) | BSS.null remainder -> Just blockno
         _                                              -> Nothing
 
+    parseRefType str =
+      case BSS.uncons str of
+        Just (typeCode, remainder)
+          | BSS.null remainder && Tar.isBuildTreeRefTypeCode typeCode
+            -> Just (refTypeFromTypeCode typeCode)
+        _   -> Nothing
+
 showIndexCacheEntry :: IndexCacheEntry -> String
-showIndexCacheEntry entry = case entry of
-   CachePackageId pkgid b -> "pkg: " ++ display (packageName pkgid)
-                                  ++ " " ++ display (packageVersion pkgid)
-                          ++ " b# " ++ show b
-   CachePreference dep     -> "pref-ver: " ++ display dep
+showIndexCacheEntry entry = unwords $ case entry of
+   CachePackageId pkgid b -> [ packageKey
+                             , display (packageName pkgid)
+                             , display (packageVersion pkgid)
+                             , blocknoKey
+                             , show b
+                             ]
+   CacheBuildTreeRef t b  -> [ buildTreeRefKey
+                             , [typeCodeFromRefType t]
+                             , show b
+                             ]
+   CachePreference dep    -> [ preferredVersionKey
+                             , display dep
+                             ]
 
 readIndexCache :: BSS.ByteString -> [IndexCacheEntry]
-readIndexCache = catMaybes . map readIndexCacheEntry . BSS.lines
+readIndexCache = mapMaybe readIndexCacheEntry . BSS.lines
 
 showIndexCache :: [IndexCacheEntry] -> String
 showIndexCache = unlines . map showIndexCacheEntry
