@@ -1,5 +1,5 @@
 {-# LANGUAGE Safe #-}
-{-# LANGUAGE CPP, ForeignFunctionInterface #-}
+{-# LANGUAGE CPP #-}
 
 -----------------------------------------------------------------------------
 -- |
@@ -17,49 +17,33 @@
 
 module System.Environment
     (
-      getArgs,            -- :: IO [String]
-      getProgName,        -- :: IO String
-      getExecutablePath,  -- :: IO FilePath
-      getEnv,             -- :: String -> IO String
-      lookupEnv,          -- :: String -> IO (Maybe String)
-#ifndef __NHC__
+      getArgs,
+      getProgName,
+      getExecutablePath,
+      getEnv,
+      lookupEnv,
+      setEnv,
+      unsetEnv,
       withArgs,
       withProgName,
-#endif
-#ifdef __GLASGOW_HASKELL__
       getEnvironment,
-#endif
   ) where
 
-import Prelude
-
-#ifdef __GLASGOW_HASKELL__
-import Foreign.Safe
+import Foreign
 import Foreign.C
-import Control.Exception.Base   ( bracket )
+import System.IO.Error (mkIOError)
+import Control.Exception.Base (bracket, throwIO)
 -- import GHC.IO
 import GHC.IO.Exception
 import GHC.IO.Encoding (getFileSystemEncoding)
 import qualified GHC.Foreign as GHC
 import Data.List
+import Control.Monad
 #ifdef mingw32_HOST_OS
 import GHC.Environment
 import GHC.Windows
 #else
-import Control.Monad
-#endif
-#endif
-
-#ifdef __HUGS__
-import Hugs.System
-#endif
-
-#ifdef __NHC__
-import System
-  ( getArgs
-  , getProgName
-  , getEnv
-  )
+import System.Posix.Internals (withFilePath)
 #endif
 
 import System.Environment.ExecutablePath
@@ -74,7 +58,8 @@ import System.Environment.ExecutablePath
 # endif
 #endif
 
-#ifdef __GLASGOW_HASKELL__
+#include "HsBaseConfig.h"
+
 -- ---------------------------------------------------------------------------
 -- getArgs, getProgName, getEnv
 
@@ -226,6 +211,8 @@ foreign import WINDOWS_CCONV unsafe "windows.h GetLastError"
 -- there is no such value.
 --
 -- For POSIX users, this is equivalent to 'System.Posix.Env.getEnv'.
+--
+-- @since 4.6.0.0
 lookupEnv :: String -> IO (Maybe String)
 #ifdef mingw32_HOST_OS
 lookupEnv name = withCWString name $ \s -> try_size s 256
@@ -238,7 +225,7 @@ lookupEnv name = withCWString name $ \s -> try_size s 256
           | otherwise  -> peekCWString p_value >>= return . Just
 
 foreign import WINDOWS_CCONV unsafe "windows.h GetEnvironmentVariableW"
-  c_GetEnvironmentVariable :: LPTSTR -> LPTSTR -> DWORD -> IO DWORD
+  c_GetEnvironmentVariable :: LPWSTR -> LPWSTR -> DWORD -> IO DWORD
 #else
 lookupEnv name =
     withCString name $ \s -> do
@@ -257,6 +244,95 @@ ioe_missingEnvVar :: String -> IO a
 ioe_missingEnvVar name = ioException (IOError Nothing NoSuchThing "getEnv"
     "no environment variable" Nothing (Just name))
 
+-- | @setEnv name value@ sets the specified environment variable to @value@.
+--
+-- On Windows setting an environment variable to the /empty string/ removes
+-- that environment variable from the environment.  For the sake of
+-- compatibility we adopt that behavior.  In particular
+--
+-- @
+-- setEnv name \"\"
+-- @
+--
+-- has the same effect as
+--
+-- @
+-- `unsetEnv` name
+-- @
+--
+-- If you don't care about Windows support and want to set an environment
+-- variable to the empty string use @System.Posix.Env.setEnv@ from the @unix@
+-- package instead.
+--
+-- Throws `Control.Exception.IOException` if @name@ is the empty string or
+-- contains an equals sign.
+--
+-- @since 4.7.0.0
+setEnv :: String -> String -> IO ()
+setEnv key_ value_
+  | null key       = throwIO (mkIOError InvalidArgument "setEnv" Nothing Nothing)
+  | '=' `elem` key = throwIO (mkIOError InvalidArgument "setEnv" Nothing Nothing)
+  | null value     = unsetEnv key
+  | otherwise      = setEnv_ key value
+  where
+    key   = takeWhile (/= '\NUL') key_
+    value = takeWhile (/= '\NUL') value_
+
+setEnv_ :: String -> String -> IO ()
+#ifdef mingw32_HOST_OS
+setEnv_ key value = withCWString key $ \k -> withCWString value $ \v -> do
+  success <- c_SetEnvironmentVariable k v
+  unless success (throwGetLastError "setEnv")
+
+foreign import WINDOWS_CCONV unsafe "windows.h SetEnvironmentVariableW"
+  c_SetEnvironmentVariable :: LPTSTR -> LPTSTR -> IO Bool
+#else
+
+-- NOTE: The 'setenv()' function is not available on all systems, hence we use
+-- 'putenv()'.  This leaks memory, but so do common implementations of
+-- 'setenv()' (AFAIK).
+setEnv_ k v = putEnv (k ++ "=" ++ v)
+
+putEnv :: String -> IO ()
+putEnv keyvalue = do
+  s <- getFileSystemEncoding >>= (`GHC.newCString` keyvalue)
+  -- IMPORTANT: Do not free `s` after calling putenv!
+  --
+  -- According to SUSv2, the string passed to putenv becomes part of the
+  -- enviroment.
+  throwErrnoIf_ (/= 0) "putenv" (c_putenv s)
+
+foreign import ccall unsafe "putenv" c_putenv :: CString -> IO CInt
+#endif
+
+-- | @unSet name@ removes the specified environment variable from the
+-- environment of the current process.
+--
+-- Throws `Control.Exception.IOException` if @name@ is the empty string or
+-- contains an equals sign.
+--
+-- @since 4.7.0.0
+unsetEnv :: String -> IO ()
+#ifdef mingw32_HOST_OS
+unsetEnv key = withCWString key $ \k -> do
+  success <- c_SetEnvironmentVariable k nullPtr
+  unless success $ do
+    -- We consider unsetting an environment variable that does not exist not as
+    -- an error, hence we ignore eRROR_ENVVAR_NOT_FOUND.
+    err <- c_GetLastError
+    unless (err == eRROR_ENVVAR_NOT_FOUND) $ do
+      throwGetLastError "unsetEnv"
+#else
+
+#ifdef HAVE_UNSETENV
+unsetEnv key = withFilePath key (throwErrnoIf_ (/= 0) "unsetEnv" . c_unsetenv)
+foreign import ccall unsafe "__hsbase_unsetenv" c_unsetenv :: CString -> IO CInt
+#else
+unsetEnv key = setEnv_ key ""
+#endif
+
+#endif
+
 {-|
 'withArgs' @args act@ - while executing action @act@, have 'getArgs'
 return @args@.
@@ -371,4 +447,3 @@ divvy str =
   case break (=='=') str of
     (xs,[])        -> (xs,[]) -- don't barf (like Posix.getEnvironment)
     (name,_:value) -> (name,value)
-#endif  /* __GLASGOW_HASKELL__ */
